package org.dbxml.core.corba;

/*
 * dbXML License, Version 1.0
 *
 *
 * Copyright (c) 1999-2001 The dbXML Group, L.L.C.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        The dbXML Group (http://www.dbxml.com/)."
 *    Alternately, this acknowledgment may appear in the software
 *    itself, if and wherever such third-party acknowledgments normally
 *    appear.
 *
 * 4. The names "dbXML" and "The dbXML Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    info@dbxml.com.
 *
 * 5. Products derived from this software may not be called "dbXML",
 *    nor may "dbXML" appear in their name, without prior written
 *    permission of The dbXML Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE DBXML GROUP OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: SymbolSerializer.java,v 1.4 2001/08/05 07:22:42 tom Exp $
 */

import org.dbxml.core.corba.db.*;

import org.dbxml.xml.*;
import org.dbxml.xml.dom.*;

import org.w3c.dom.*;

/**
 * SymbolSerializer is a utility class for managing SymbolTables in
 * the server context of Wire Compression.
 */

public final class SymbolSerializer {
   private static final byte[] EmptyBytes = new byte[0];
   private static final SymbolTableSymbols hcSyms = SymbolTableSymbols.getInstance();
   private static final DocumentImpl elemFactory = new DocumentImpl();
   
   private SymbolTable syms = null; // The Collection's SymbolTable
   private long lastMod = 0;        // Last time we caught a SymbolTable modification
   private byte[] symBytes = null;  // Stores a byte representation of the SymbolTable
   
   public SymbolSerializer(SymbolTable syms) {
      this.syms = syms;
      elemFactory.setSymbols(syms);
   }

   /**
    * getSymBuffer returns a new EncodedBuffer that includes a
    * reference to the SymbolTable's byte array image for wire
    * transmission.  Depending on whether the client is up to
    * date, convertFromDocument will remove this reference.
    *
    * @return A new EncodedBuffer
    */
   public EncodedBuffer getSymBuffer() {
      long lm = syms.getLastModified();
      if ( lm > lastMod ) {
         synchronized ( syms ) {
            Element elem = syms.streamToXML(elemFactory);
            try {
               symBytes = DOMCompressor.Compress(elem, hcSyms);
               lastMod = lm;
            }
            catch ( Exception e ) {
               org.dbxml.Debug.println("Problem compressing Symbol Table!  Very Bad!");
            }
         }
      }
      return new EncodedBuffer(lm, symBytes, null);
   }
   
   /**
    * convertFromDocument converts a DOM Document into an
    * EncodedBuffer that, depending on the time stamp, possibly
    * includes a current image of the managed Collection's Symbol
    * Table.
    *
    * @param doc The Document to Convert
    * @param stamp The client's last modified stamp
    * @return The EncodedBuffer
    */
   public EncodedBuffer convertFromDocument(Document doc, long stamp) {
      EncodedBuffer buf = getSymBuffer();
      byte[] docBytes = ((CompressedDocument)doc).getDataBytes();
      buf.buf = docBytes;
      if ( buf.stamp == stamp )
         buf.sym = EmptyBytes;
      return buf;
   }
   
   /**
    * getSymbols returns the Symbol Table being managed by this
    * Serializer.
    *
    * @return The Symbol Table
    */
   public SymbolTable getSymbols() {
      return syms;
   }
   
   /**
    * getLastModified returns a time stamp of the last server-side
    * modified of the Symbol Table.  This is used to determine
    * whether or not to regenerate the byte stream, and whether
    * the client will need a new copy of the Symbol Table.
    *
    * @return Last modified stamp
    */
   public long getLastModified() {
      return lastMod;
   }
}

